package heap;
import java.util.*;
import util.*;
import code.*;

/** Represents a Symbolic Shape Graph. As in the theory, a SSG is defined by his counters, vertices, and roots with the bindings between them.
 * This class is Cloneable to make easy copying a SSG (which is one of the main actions perfomed during CA construction).
 * The Equivalent interface is here to implement alpha-equivalence between two SSG.
 * @author Swann.Perarnau@imag.fr
 *  
 */
public class Heap implements Cloneable,Equivalent{
	
		/**Contains the bindings between Roots and Vertices */
    	private Map 	rootBindings;
    	
    	/**Contains the successor mapping between Vertices. A Vertex V points to a Vertex W if this map contains (V,W). */
    	private Map		verticesSuccessor;
    	
    	/**List of Roots */
        private List 	roots;
        
        /**List of Vertices */
        private List 	vertices;
        
        /**List of counters. Used when creating a new vertex to ensure that the counter we create is not already used. */
    	private List 	counters;
        

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//    	Constructors
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////    	
    	
    	
    	/**Create an empty SSG */
        public Heap() {
            rootBindings = new HashMap();
            verticesSuccessor = new HashMap();
            roots = new LinkedList();
            vertices = new LinkedList();
            counters = new LinkedList();
        }
        
        /**Create a SSG with the same content that another one.
         * @param h the SSG to fill the new one with
         *  */
        public Heap(Heap h) {
            rootBindings = new HashMap(h.rootBindings);
            verticesSuccessor = new HashMap(h.verticesSuccessor);
            roots = new LinkedList(h.roots);
            vertices = new LinkedList(h.vertices);
            counters = new LinkedList(h.counters);
        }

        

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Getters
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
          
        /**Returns the list of all Vertices of this SSG.*/
        public List getVertices(){
        	return vertices;
        }
        
        /**Returns the list of all Roots of this SSG.*/
        public List getRoots(){
        	return roots;
        }
        
        /**Returns the list of all counters of this SSG.*/
        public List getCounters(){
        	return counters;
        }
        
        /**Returns the list of Roots which point to a Vertex.*/
        public List getRoots(Vertex v){
        	List l = new LinkedList();
        	if(v == null)
        		return l;
        	Set e = rootBindings.entrySet();
            Iterator i = e.iterator();
            while(i.hasNext())
            {
                Map.Entry tmp = (Map.Entry) i.next();
                if(v.equals(tmp.getValue()))
                    l.add(tmp.getKey());                
            }
            return l;
        }
        
        /**Returns the Vertex pointed by a Root or null.*/
        public Vertex getVertex(Root r) {
            return (Vertex) rootBindings.get(r);
        }
        
        /**Returns this SSG vertex equals to the argument.*/
        public Vertex getVertex(Vertex v) {
            int i = vertices.indexOf(v);
            if(i == -1)
            	return null;
            else
            	return (Vertex)vertices.get(i);
        }
        
        /**Returns the successor of a Vertex or null. */
        public Vertex getSuccessor(Vertex v)
        {
        	return (Vertex) verticesSuccessor.get(v);
        }
        
        /**Returns the number of vertices in this SSG.*/
        public int getNbVertices()
        {
        	return vertices.size();
        }
        
        /**Returns the number of roots in this SSG.*/
        public int getNbRoots()
        {
        	return roots.size();
        }      
        
        /**Returns the root of corresponding name.*/
        public Root getRoot(String s)
        {
        	Root r = new Root(s);
        	int i = roots.indexOf(r);
        	if(i == -1)
        		return null;
        	else
        		return (Root)roots.get(i);
        }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Add/Remove methods
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        
        /**Adds a Root to this SSG.
         * @return true if the Root has been added, false if there is an error.
         */
        public boolean add(Root r)
        {
        	if(roots.contains(r))
        		return false;
        	else
        	{
        		return roots.add(r);        		
        	}
        }
        
        /** Adds a Vertex to this SSG. Update this SSG counter list accordingly.
         * @return true if the Vertex has been added, false if there is an error.
         */
        public boolean add(Vertex v)
        {
        	if(vertices.contains(v))
        		return false;
        	else
        	{
        		counters.add(v.counter());
        		return vertices.add(v);
        	}
        }
        
        /**Removes the Root from this SSG. Updates the bindings.
         * @return true if removing is done.
         */
        public boolean remove(Root r) {
        	if(r == null)
        		return false;
        	
            rootBindings.remove(r);
            roots.remove(r);
            return true;
        }
        
        /**Removes the Vertex from this SSG. Updates the bindings and counters.
         * @return true if removing is done.
         */
        public boolean remove(Vertex v)
        {
        	if(v == null)
        		return false;
        	
        	List roots = getRoots(v);
        	Iterator i = roots.iterator();
        	while(i.hasNext())
        	{
        		Root r = (Root) i.next();
        		unlink(r,v);
        	}
        	
        	Vertex succ = getSuccessor(v);
        	if(succ != null)
        		unlink(v,succ);
        	
        	vertices.remove(v);
        	counters.remove(v.counter());
        	return true;
        }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Link management
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
         
        /**Checks if a Root is linked to a Vertex.*/
        public boolean isLinked(Root r)
        {
        	if(r == null)
        		return false;
        	
        	return rootBindings.containsKey(r);
        }
        
        /**Checks if a Root is linked to a Vertex.*/
        public boolean isRooted(Vertex v)
        {
        	if(v == null)
        		return false;
        	
        	return rootBindings.containsValue(v);
        }
        
        /**Checks if a Vertex has a succesor.*/
        public boolean hasSuccessor(Vertex v)
        {
        	if(v == null)
        		return false;
        	
        	return verticesSuccessor.containsKey(v);
        }
        
        /**Links a Root to a Vertex. If the Vertex is null or Root is already linked, no action is performed.*/ 
        public void link(Root r,Vertex v) {
        	if(r != null && v != null)
        	{
        		if(!isLinked(r))
        		{
        			if(!roots.contains(r))
        				add(r);

        			if(!vertices.contains(v))
        				add(v);

        			rootBindings.put(r,v);
        		}
        	}            		            
        }
        
        /**Links a Vertex to a successor. If the Vertex is null or already linked, no action is performed.*/ 
        public void link(Vertex v, Vertex succ) {
        	if(v != null && succ != null)
        	{
        		if(!hasSuccessor(v))
        		{
        			if(!vertices.contains(v))
        				add(v);

        			if(!vertices.contains(succ))
        				add(succ);

        			verticesSuccessor.put(v,succ);
        		}
        	}            		            
        }
        
        /**Removes the link between a Root and a Vertex.*/
        public void unlink(Root r,Vertex v) {
			if (r != null && v != null) {
				if (rootBindings.containsKey(r)
						&& rootBindings.containsValue(v)
						&& rootBindings.get(r).equals(v))
					rootBindings.remove(r);
			}            
        }
        
        /**Removes the link between a Vertex and his successor.*/
        public void unlink(Vertex v,Vertex succ) {
			if (v != null && succ != null) {
				if (verticesSuccessor.containsKey(v)
						&& verticesSuccessor.containsValue(succ)
						&& verticesSuccessor.get(v).equals(succ))
					verticesSuccessor.remove(v);
			}            
        }
        

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Huge transformations
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        
        
        /**Normalizes a SSG : a graph is in normal form if each of its vertices is either rooted or successor of at least two vertices.
         * Return a set of instructions on the counters, representing the operations that have occured on the counters.
         */
        public CodeNode normalize()
        {
        	boolean toremove = false;
        	CodeNode n = null;
        	CodeNode m = null;
        	/* delete vertices which are not cutpoints */
        	Iterator i = vertices.iterator();
        	while(i.hasNext())
        	{
        		Vertex v = (Vertex) i.next();
        		if(!isRooted(v)) /* no root */
        		{
        			/* how many predecessors ? */
        			int nbPreds = 0;
        			Vertex w = null,x = null; // x is the predecessor if there is only one.
        			Iterator j = vertices.iterator();
        			while(j.hasNext())
        			{
        				w = (Vertex) j.next();
        				if(v.equals(getSuccessor(w)))
        				{
        					nbPreds++;
        					if(x == null)
        						x = w;        					
        				}
        			}
        			if(nbPreds < 2)
        			{
        				if(x != null) // if false the vertex is not accessible from the program
        				{
        					m = CodeNode.actions_xxy(x.counter(), v.counter());
        					if(n == null)
        						n = m;
        					else
        						n = CodeNode.slist(n, m); // building list of actions
        					Vertex vNext = getSuccessor(v);
        					unlink(x,v);
        					link(x,vNext);
        				}
        				toremove = true;
        			}
        		}
        		if(toremove)
        		{
        			toremove = false;
        			counters.remove(v.counter());
        			i.remove();
        		}
        	} 
        	return n;
        }
        
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Ligth transformations // Semantical actions on the graph
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
         
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := y".*/
        public void asignXToY(String x,String y)
        {
        	Root ry = getRoot(y);
			Vertex vy = getVertex(ry);
			if(vy == null) // y -> nil
			{
				Root rx = getRoot(x);
				Vertex vx = getVertex(rx);
				if(vx == null) // x -> nil
				{
					;
				}
				else
				{
					unlink(rx, vx);					
				}
			}
			else
			{
				Root rx = getRoot(x);
				Vertex vx = getVertex(rx);
				if(vx == null) // x -> nil
				{
					link(rx, vy);
				}
				else
				{
					unlink(rx, vx);					
					link(rx, vy);
				}
			}
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := null".*/
        public void asignXToNull(String x)
        {
			Root rx = getRoot(x);
			Vertex v = getVertex(rx);
			if(v == null) // x -> nil
			{
				;
			}
			else
			{
				unlink(rx, v);				
			}
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := new".*/
        public void asignXToNew(String x)
        {
			// erase previous value of x
			Root rx = getRoot(x);
			Vertex v = getVertex(rx);
			if(v != null)
			{
				unlink(rx, v);				
			}
			
			// create new vertice
			Vertex n = new Vertex(uniqueCounter());
			add(n);
			link(rx, n);
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := x.next" and with the following condition true : counter(x) == 1.*/
        public void asignXToXNextWithCounterXEquals1(String x)
        {
        	Root rx = getRoot(x);
        	Vertex vx = getVertex(rx);
        	if(vx != null)
        	{
        		Vertex vNext = getSuccessor(vx);
        		unlink(rx,vx);
        		link(rx,vNext);        		
        	}
        	// else ERROR TODO : check that, I don't remember if there's a problem here
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := x.next" and with the following condition true : counter(x) > 1.*/
        public void asignXToXNextWithCounterXGreaterThan1(String x)
        {
        	Root rx = getRoot(x);
        	Vertex vx = getVertex(rx);
        	if(vx != null)
        	{
        		// we create a new vertex and add it to this SSG
        		Vertex v = new Vertex(uniqueCounter());
        		add(v);
        		
        		// we link this vertex to the successor of vx, and unlink vx with.
        		link(v,getSuccessor(vx));        		
        		unlink(vx,getSuccessor(vx));
        		
        		// then we link vx with his new successor and move the root
        		link(vx,v);
        		unlink(rx,vx);
        		link(rx,v);        		
        	}
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := x".*/
        public void asignXToX(String x) { ; }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := y.next" and with the following condition true : counter(y) == 1.*/
        public void asignXToYNextWithCounterYEquals1(String x,String y)
        {
        	Root ry = getRoot(y);
			Vertex vy = getVertex(ry);
			if(vy != null)
			{
				Root rx = getRoot(x);
				unlink(rx,getVertex(rx));
				link(rx,getSuccessor(vy));				
			}        	
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x := y.next" and with the following condition true : counter(y) > 1.*/
        public void asignXToYNextWithCounterYGreaterThan1(String x,String y)
        {
        	Root ry = getRoot(y);
			Vertex vy = getVertex(ry);
			if(vy != null)
			{
				Root rx = getRoot(x);
				Vertex v = new Vertex(uniqueCounter());
				add(v);
				link(v,getSuccessor(vy));
				
				unlink(vy,getSuccessor(vy));
				link(vy,v);
				
				unlink(rx,getVertex(rx));
				link(rx,v);
				
			}
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x.next := y".*/
        public void asignXNextToY(String x,String y)
        {
        	Root rx = getRoot(x);
        	Vertex vx = getVertex(rx);
        	if(vx != null)
        	{
        		Root ry = getRoot(y);
        		unlink(vx,getSuccessor(vx));
        		link(vx,getVertex(ry));
        	}
        }
        
        /**Makes the SSG tranformations corresponding to an assignement between roots "x.next := null".*/
        public void asignXNextToNull(String x)
        {
        	Root rx = getRoot(x);
        	Vertex vx = getVertex(rx);
        	if(vx != null)
        	{
        		unlink(vx,getSuccessor(vx));
        	}
        }


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Utils
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		


		public int hashCode() {
			final int PRIME = 31;
			int result = 1;
			result = PRIME * result + ((counters == null) ? 0 : counters.hashCode());
			result = PRIME * result + ((rootBindings == null) ? 0 : rootBindings.hashCode());
			result = PRIME * result + ((roots == null) ? 0 : roots.hashCode());
			result = PRIME * result + ((vertices == null) ? 0 : vertices.hashCode());
			result = PRIME * result + ((verticesSuccessor == null) ? 0 : verticesSuccessor.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			final Heap other = (Heap) obj;
			if (counters == null) {
				if (other.counters != null)
					return false;
			} else if (!counters.equals(other.counters))
				return false;
			if (rootBindings == null) {
				if (other.rootBindings != null)
					return false;
			} else if (!rootBindings.equals(other.rootBindings))
				return false;
			if (roots == null) {
				if (other.roots != null)
					return false;
			} else if (!roots.equals(other.roots))
				return false;
			if (vertices == null) {
				if (other.vertices != null)
					return false;
			} else if (!vertices.equals(other.vertices))
				return false;
			if (verticesSuccessor == null) {
				if (other.verticesSuccessor != null)
					return false;
			} else if (!verticesSuccessor.equals(other.verticesSuccessor))
				return false;
			return true;
		}

		/**Display a SSG */
        public String toString()
        {
        	String s = new String("\n\nHeap : \n Vertices : \n");
        	
        	Iterator i = vertices.iterator();
        	while(i.hasNext())
        	{
        		Vertex v = (Vertex) i.next();
        		Vertex next = getSuccessor(v);
        		s += v.toString() + "->" + ((next == null)?"null":(""+next.fileid()));
        	}
        	
        	s += " Roots : \n";
        	Iterator j = roots.iterator();
        	while(j.hasNext())
        	{
        		Root r = (Root) j.next();
        		s += r.toString() + " ";
        	}
        	s += "\n";
        	s += " rootBindingsping : \n";
        	Set e = rootBindings.entrySet();
        	Iterator k = e.iterator();
        	while(k.hasNext())
        	{
        		Map.Entry tmp = (Map.Entry)k.next();
        		s += ((Root)tmp.getKey()).name() + " --> " + ((Vertex)tmp.getValue()).fileid() + "\n";
        	}
        	s += "\n\n";
        	return s;
        }
        
        
        /**Searches for a Vertex with a specified FileId. */
        public Vertex findFileId(int fileid)
        {
        	Iterator i = vertices.iterator();
            while(i.hasNext())
            {
                Vertex tmp = (Vertex) i.next();
                if(tmp.fileid() == fileid)
                    return tmp;                
            }
            return null;
        }
        
        /**Generates a counter name not already used in this SSG.*/
        public String uniqueCounter()
        {
        	String s = "c";
        	int i = 0;
        	while(counters.contains(s+i))i++;
        	return s+i;
        }
        
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Implemeted Interfaces
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				
		/** Alpha-equivalence. Assume that the two SSG are normalized ! 
		 * @return a Map (this.vertex,o.vertex) of equivalence
		 */
        public Object isEquivalentTo(Object o) 
        {
        	if(!(o instanceof Heap))
        		return null;
        	else
        	{
        		Heap h = (Heap) o;
        		// they are normalized so they have the same number of vertices
        		if(h.getNbVertices() != this.getNbVertices())
        			return null;
        		
        		// they can't be equivalent if some roots are missing
        		if(! (h.roots.containsAll(this.roots) && this.roots.containsAll(h.roots)))
        			return null;
        		
        		Map m = new HashMap();
        		/* put all the vertices and null equivalence */
        		Iterator i = vertices.iterator();
        		while(i.hasNext())
        		{
        			Vertex v = (Vertex) i.next();
        			m.put(v, null);
        		}
        		
        		List l = new LinkedList();
        		/* take care of rooted vertices first */
        		i = roots.iterator();
        		while(i.hasNext())
        		{
        			Root r = (Root) i.next();
        			Vertex v = getVertex(r);
					if(v == null)
					{
						// r must point on null in the other SSG
						if(h.getVertex(r) != null)
							return null;
					}
					else
					{
						// we check that the two vertices are equivalent (pointed by same roots)
						Vertex v_prime = h.getVertex(r);
						if(v_prime == null)
							return null;
						
						List v_roots = getRoots(v);
						List v_prime_roots = h.getRoots(v_prime);
						
						if(!(v_roots.containsAll(v_prime_roots) && v_prime_roots.containsAll(v_roots)))
								return null;
						
						m.put(v,v_prime);
						if(!l.contains(v))
							l.add(v);        			
					}
        		}	
        			
        		/* now we compare vertices for which the predecessor is known */	
        		ListIterator a = l.listIterator(l.size()); // fathers list
        		while(a.hasPrevious())
        		{
        			Vertex v = (Vertex)a.previous();
        			Vertex v_next = getSuccessor(v);
        			Vertex v_prime = (Vertex) m.get(v);
        			if(v_prime == null)
        				throw new RuntimeException("At heap.isEquivalentTo : Missing vertex impossible");
        			
        			Vertex v_prime_next = h.getSuccessor(v_prime);
        			if(v_next == null) // no son
        			{
        				if(v_prime_next != null)
        					return null;
        				a.remove();
        			}
        			else if(v_prime_next != null) // are the sons equivalent ?
        			{
        				if(m.containsValue(v_prime_next)) // there's already a mapping
        				{
        					if(m.get(v_next) == null || !m.get(v_next).equals(v_prime_next)) // not a good one
        						return null;
        					else
        					{
        						a.remove();
        					}
        				}
        				else
        				{
        					if(m.get(v_next) == null) // we find a new mapping
            				{
            					m.put(v_next, v_prime_next);
            					a.remove();
            					if(!l.contains(v_next))
            							a.add(v_next);
            				}
            				else // is the mapping already contained good ?
            				{
            					if(!m.get(v_next).equals(v_prime_next))
            						return null;
            					a.remove();
            					if(!l.contains(v_next))
            						a.add(v_next);
            				}
        				}
        			}	
        			else
        			{
        				return null;
        			}
        		}
        		if(m.containsValue(null)) // there a node not mapped
        			return null;
        		
        		return m;
        	}
        }        			
        
        
       
        /**Return an exact clone of this SSG */
        public Object clone()
        {
        	Heap h = new Heap();
        	h.counters.addAll(counters);
        	h.roots.addAll(roots);
        	
        	/*We must clone the vertices too*/
        	Iterator j = vertices.iterator();
        	while(j.hasNext())
        	{
        		Vertex v = (Vertex)j.next();
        		h.vertices.add(new Vertex(v));
        	} 
        	
        	Iterator l = h.vertices.iterator();
        	while(l.hasNext())
        	{
        		Vertex v = (Vertex)l.next();
        		Vertex oldv = getVertex(v);
        		if(oldv == null)
        			throw new RuntimeException("At Heap.Clone : Missing vertex impossible !");
        		Vertex next = getSuccessor(oldv);
        		h.link(v,h.getVertex(next));
        	}
        	
        	Iterator k = h.roots.iterator();
        	while(k.hasNext())
        	{
        		Root r = (Root)k.next();
        		Vertex v = getVertex(r);
        		if(v != null)
        		h.link(r, h.findFileId(v.fileid()));
        	}     	
        	return h;
        }

}
